#!/bin/perl
################################################################################
# Created       : Martin Foster
# Modified      : 18-Oct-2003 
################################################################################
#
# Config Realm - Script part of Ethereal Realms, designed for supervisors
#                to control realms
# Copyright (C) 2000-2004  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 29-26503 TWP RD 511
#                 Spruce Grove, Alberta
#                 T7Y 1G4
#
#################################################################################

use CGI qw(:standard);						# Common gateway interface
use CGI::Carp qw(fatalsToBrowser);				# CGI Error logs
use strict;							# Strict variable enforcement

use eRealms::Audit;						# Auditing information
use eRealms::Comm;						# Comm panel
use eRealms::Database;						# Database handler
use eRealms::Filter;						# Filtering handle
use eRealms::Login;						# Login functionality
use eRealms::Mail;						# Mail handler
use eRealms::Option;						# Option handler
use eRealms::Param;						# Parameter control
use eRealms::Pull;						# Post retrieval
use eRealms::Table;						# Table handler
use eRealms::Template;						# Template handler
use eRealms::Weather;						# Weather handler
use eRealms::Who;						# Who is on


################################################################################
# Data Members
################################################################################
my $cgi;							# Common gateway interface handle
my $database;							# Database handle
my $login;							# Login handle
my $param;							# Parameter handle
my $pull;							# Post retreival
my $table;							# Table handle
my $tmpl;							# Template handle

my $admin;							# Administrative override
my $access;							# Realm access
my $acl;							# Access control list
my $host = 15;							# Requirement to be over for host
my $url;							# Self referencing link
my $verdict;							# Verdict on bans

my @acl;							# Captured ACL

my %args;							# Default arguments
my %config;							# Configuration hash


################################################################################
# Program Area
################################################################################

	# Initial handle
	$cgi      = new CGI;	
	$database = new eRealms::Database();
	$pull     = new eRealms::Pull();

	# Connect and fetch
	$database->Connect();


	# Pre-capture
	@acl = $cgi->param('MODACL');


	# Parameter handling
	$param = new eRealms::Param($database, $cgi);
	$param->GetParam();

	# Fetch user
	$url = $cgi->url(-absolute=>1);

	# Fetch hash
	$database->GetHashConfiguration(\%config);

	$table    = new eRealms::Table($cgi);
	$tmpl     = new eRealms::Template(\%config);


	# Document header
	print $cgi->header();
	$database->DocumentGetHeader();

	# Login handle
	$login = new eRealms::Login($database, $cgi, $param);

	# Authenthication
	if ($login->GetVerificationSuper())
	{
		# Display title
		print "$config{'TagTitle'}\n<BR>\n";

		# Fetch Posts
		$pull->AdminPosts($database, $cgi, $param);

		# Realm selection
		unless (defined($param->{'ROOM'}))
		{ 
			RealmSelect($database, $cgi, $param, $table, $tmpl, \%config); 
		}

		# Realm handling
		else
		{ 
			# Prevent bans from use
			($access)  = $database->DataGetRealmAccess($param->{'ROOM'});
			($acl)     = $database->DataGetRealmACL($param->{'USER'}, $param->{'ROOM'});
			($verdict) = $database->DataGetBannedStatus($param->{'USER'}, $param->{'ROOM'});

			# Administrative access
			($admin)   = $database->DataGetAdmin($param->{'USER'}); 


			# Only show if not banned
			unless (((($access eq 'relaxed') && (defined($verdict)))
			      or (($access eq 'restricted') && (defined($verdict)))
			      or (($access eq 'list') && (!defined($verdict))))
			     and ($admin eq 'no'))
			{
				# Generate arguments
				%args = (
					USER  => $param->{'USER'},
					CRYPT => $param->{'CRYPT'}
				 );


				# Generate inlines
				# Back button
				my ($iback) = $param->EmbedInline(%args); 


				# Add parameter
				$args{'ROOM'} = $param->{'ROOM'};

				# Regular inlines
				my ($ibann) = $param->EmbedInline(%args, BANN => 'True');
				my ($icnon) = $param->EmbedInline(%args, CANON => 'True');
				my ($idsgn) = $param->EmbedInline(%args, DESIGN => 'True');
				my ($ifntc) = $param->EmbedInline(%args, FONT => 'True');
				my ($iforc) = $param->EmbedInline(%args, FORCAST => 'True');
				my ($ipics) = $param->EmbedInline(%args, GRAPH => 'True');
				my ($igumm) = $param->EmbedInline(%args, GUMMY => 'True');
				my ($ilocl) = $param->EmbedInline(%args, LOCALE => 'True');
				my ($imods) = $param->EmbedInline(%args, MODS => 'True');
				my ($itopc) = $param->EmbedInline(%args, TOPIC => 'True');
				my ($iwarn) = $param->EmbedInline(%args, WARN => 'True');
				my ($iword) = $param->EmbedInline(%args, WORD => 'True');
				my ($ipost) = $param->EmbedInline(%args, GETPOSTS => 'True');
				my ($irelm) = $param->EmbedInline(%args);


				# Display Links
				# Normal supervisory display
				if ($acl > $host)
				{
					# Display every option
				 	print "<CENTER><A HREF=\"$url$iback\">$config{TagBack}</A> ",
					      "<A HREF=\"$url$irelm\">$config{TagRealm}</A> ",
					      "<A HREF=\"$url$ibann\">$config{TagBann}</A> ",
					      "<A HREF=\"$url$itopc\">$config{TagTopic}</A> ",
					      "<A HREF=\"$url$icnon\">$config{TagCanon}</A> ",
					      "<A HREF=\"$url$idsgn\">$config{TagDesign}</A> ",
					      "<A HREF=\"$url$ifntc\">$config{TagFont}</A> ",
					      "<A HREF=\"$url$iforc\">$config{TagForcast}</A> ",
					      "<A HREF=\"$url$igumm\">$config{TagGummy}</A> ",
					      "<A HREF=\"$url$ilocl\">$config{TagLocale}</A> ",
					      "<A HREF=\"$url$imods\">$config{TagMods}</A> ",
					      "<A HREF=\"$url$ipics\">$config{TagPics}</A> ",
					      "<A HREF=\"$url$ipost\">$config{TagPosts}</A> ",
					      "<A HREF=\"$url$iwarn\">$config{TagWarn}</A>",
					      "<A HREF=\"$url$iword\">$config{TagWord}</A></CENTER>\n<BR><BR>\n";
				}


				# Moderative 'Me Too!'
				else
				{
					# General
				 	print "<CENTER><A HREF=\"$url$iback\">$config{TagBack}</A> ";

					# Storyteller
					if (($acl - 8) >= 0)
					{
						# Remove amount
						$acl -= 8;

						# Options
						print "<A HREF=\"$url$icnon\">$config{TagCanon}</A> ",
						      "<A HREF=\"$url$igumm\">$config{TagGummy}</A> ",
						      "<A HREF=\"$url$ilocl\">$config{TagLocale}</A> ";
					}

					# Jailer
					if (($acl - 4) >= 0)
					{
						# Remove amount
						$acl -= 4;

						# Options
						print "<A HREF=\"$url$ibann\">$config{TagBann}</A> ",
						      "<A HREF=\"$url$iword\">$config{TagWord}</A> ";
					}

					# Designer
					if (($acl - 2) >= 0)
					{
						# Remove amount
						$acl -= 2;

						# Options
						print "<A HREF=\"$url$idsgn\">$config{TagDesign}</A> ",
						      "<A HREF=\"$url$ifntc\">$config{TagFont}</A> ";
					}

					# Archivist
					if (($acl - 1) >= 0)
					{
						# One and only option
						print "<A HREF=\"$url$ipost\">$config{TagPosts}</A>";
					}


					# Everyone gets to play
					print "<A HREF=\"$url$iwarn\">$config{TagWarn}</A>";

					# Finish
					print "</CENTER>\n<BR><BR>\n";
				}


				# Realm information
				RoomWrite($database, $cgi, $param, \%config, $access, $admin, \@acl);

				# Bannings
				if (defined($param->{'BANN'}))
				{
					RoomBan($database, $cgi, $param, $table, $tmpl, \%config);	
				}


				# Canon handles
				elsif (defined($param->{'CANON'}))
				{
					RoomCanon($database, $cgi, $param, $table, $tmpl, \%config);
				}

				# Realm Design
				elsif (defined($param->{'DESIGN'}))
				{
					RoomDesign($database, $cgi, $param, $table, $tmpl, \%config);
				}

				# Font handler
				elsif (defined($param->{'FONT'}))
				{
					FontHandler($database, $cgi, $param, $table, $tmpl, \%config);
				}

				# Gummy handler
				elsif (defined($param->{'GUMMY'}))
				{
					GummyHandler($database, $cgi, $param, $table, $tmpl, \%config);	
				}

				# Locale handler
				elsif (defined($param->{'LOCALE'}))
				{
					LocaleHandler($database, $cgi, $param, $table, $tmpl, \%config);	
				}

				# Weather Forcasts
				elsif (defined($param->{'FORCAST'}))
				{
					Forcasts($database, $cgi, $param, $table, $tmpl, \%config) if ($acl > $host);
				}

				# Picture Restriction
				elsif (defined($param->{'GRAPH'}))
				{
					RoomPicture($database, $cgi, $param, $table, $tmpl, \%config) if ($acl > $host);
				}

				# Moderator Handling
				elsif (defined($param->{'MODS'}))
				{
					RoomMods($database, $cgi, $param, $table, $tmpl, \%config) if ($acl > $host);
				}
			

				# Board Topics
				elsif (defined($param->{'TOPIC'}))
				{
					BoardTopic($database, $cgi, $param, $table, $tmpl, \%config);
				}

				# Word Restrictions
				elsif (defined($param->{'WARN'}))
				{
					RoomWarn($database, $cgi, $param, $table, $tmpl, \%config);	
				}

				# Word Restrictions
				elsif (defined($param->{'WORD'}))
				{
					RoomWord($database, $cgi, $param, $table, $tmpl, \%config);	
				}


				# Realm information
				else
				{
					# Default for supervisors
					if ($acl > $host)
					{
						# Allow them to see what is necessary
						RoomView($database, $cgi, $param, $table, $tmpl, \%config, $admin);	
					}
					else
					{
						# Display message
						print $config{'MsgSelect'}, "\n";
					}
				}
			}
		}
	}

	$database->DocumentGetFooter();


################################################################################
# Sub-Routines
################################################################################

#####################
# Message Board Topic
#
# Allows for the addition and removal of topics in relation to their use in
# the message boards integrated with the system itself.

sub BoardTopic
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $table    = shift;					# Table handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash

	my $descr;						# Detail/description
	my $option;						# Options menu
	my $topics;						# List of topics

	my @topics;						# List of topics	


	#####################
	# program area

	# Title
	$tmpl->ShowMain($config->{'TagTopcTitle'});

	# Details screen
	if ((defined($param->{$config->{'TxtSubmitView'}}))
	 || (defined($param->{'TEDIT'})))
	{
		# Cleanup
		$param->Cleanup('TEDIT');

		# Pull details
		($descr) = $database->DataGet("SELECT BoardDescription
			FROM BoardTopic
			WHERE BoardTopic=?
			  AND RealmName=?",

		 $param->{'TOPIC'},
		 $param->{'ROOM'}
		);	


		# Display
		# Beginning of form
		print $cgi->start_form();
		print $param->EmbedNormal($param->Flat(), TOPICS=>'True', TOPIC=>$param->{'TOPIC'}), "\n";

		# Table Top
		$table->MakeTop();

		# Edit
		$table->MakeSingle($tmpl->PassSub($config->{'TagTopcEdit'}));
		$table->MakeSingle($cgi->textarea(-name=>'TEDIT',-default=>$descr,-rows=>3,-columns=>50,-wrap=>'SOFT'));
		$table->MakeBlank();

		# Buttons
		$table->MakeValid(
		  $cgi->submit($config->{'TxtSubmitAlter'}),
		  $cgi->reset($config->{'TxtResetButton'})
		);
		$table->MakeBottom();

		# End of form
		print $cgi->end_form(), "\n";
	}

	# Default display
	else
	{
		# Retreive list of banned and suspects
		$database->GetListTopics(\@topics, $param->{'ROOM'});


		# Create widget
		$topics = $cgi->scrolling_list('TOPIC',  \@topics, $topics[0], 10);
		$option = $cgi->submit($config->{'TxtSubmitDelete'}) . '&nbsp;'
		  . $cgi->checkbox(-name=>'CONFIRM', -label=>'') . '<BR><BR>'
		  . $cgi->submit($config->{'TxtSubmitView'}); 


		# HTML
		# Beginning of form
		print $cgi->start_form();
		print $param->EmbedNormal($param->Flat(), TOPICS=>'True'), "\n";

		# Table Top
		$table->MakeTop();

		# Remove
		$table->MakeSingle($tmpl->PassSub($config->{'TagTopcRem'}));
		$table->MakeValid($topics, $option, 'middle');
		$table->MakeBlank();

		# Addition
		$table->MakeSingle($tmpl->PassSub($config->{'TagTopcAdd'}));
		$table->MakeValid($cgi->textfield('TADD', '', 15, 25), $cgi->submit($config->{'TxtSubmitAddition'}));
		$table->MakeBottom();

		# End of form
		print $cgi->end_form(), "\n";
	}
}

#####################
# Font Finder
#
# Makes the ammusing task of determining exactly what to do with what is has
# been given.  If to add or remove the value from the list et cetera.

sub FontFinder
{
	#####################
	# Data members
	my $cgi    = shift;					# CGI handle
	my $param  = shift;					# Parameter handle
	my $array  = shift;					# Array to embed
	my $realm  = shift;					# Realms information 
	my $name   = shift;					# Compare value

	my $value;						# Discovered value
	my @realm;						# Realms choices


	#####################
	# Data members

	# Initial display
	unless (defined($param->{'FONTCTRL'}))
	{
		# Only perform if one can compare
		if (defined($realm))
		{
			# Seperate into useable objects
			@realm = split(/:/, $realm);
	
			# Search
			foreach $value (@realm)
			{		
				# If matched
				if ($value eq $name)
				{
					# Remove space
					$name =~ s/\s//gs;

					# For checked
					return $cgi->checkbox(-name=>$name, -label=>'', -checked=>'ON');
				}
			}
		}

		# Nothing defined assume defaults
		else
		{
			# Remove spaces
			$name =~ s/\s//gs;

			# For checked
			return $cgi->checkbox(-name=>$name, -label=>'', -checked=>'ON');
		}
	}

	# Parameter level handling
	else
	{
		# Assign value
		$value = $name;

		# Remove spaces
		$name =~ s/\s//gs;

		
		# Array handling
		if (defined($param->{$name}))
		{
			# Push onto array
			push(@{$array}, $value);

			# For checked
			return $cgi->checkbox(-name=>$name, -label=>'', -checked=>'ON');
		}
	} 

	# Return false
	return $cgi->checkbox(-name=>$name, -label=>'');
}

#####################
# Font Handler
#
# Mostly based on the same code as the administrative script, this one will 
# allow hosts to actually make a selection of which they wish to allow for
# use.

sub FontHandler
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter reference
	my $table    = shift;					# Table handler
	my $tmpl     = shift;					# Font handler
	my $config   = shift;					# Config hash

	my $audit;						# Auditing handle
	my $res;						# Results			
	my $statement;						# Database statement

	my $ctype;						# Current working type
	my $name;						# Font name
	my $type;						# Font type

	my $rcolor;						# Realms font color
	my $rface;						# Realms font face
	my $rsize;						# Realms font size
	
	my @rcolor;						# Realms font color
	my @rface;						# Realms font face
	my @rsize;						# Realms font size


	#####################
	# Program area

	# Auditing handle
	$audit = new eRealms::Audit($database, $cgi, $param);


	# Retreive realms choices
	($rcolor, $rface, $rsize) = $database->DataGetRealmFont($param->{'ROOM'});	


	# Query management
	# Prepare and execute
	$database->Pull(\$statement, "SELECT 
		 FontName AS \"FontName\",
		 FontType AS \"FontType\",
		 FontCode AS \"FontCode\"
		FROM  Font
		ORDER BY FontType, FontCode");


	# Form handle
	print $cgi->start_form();
	print $param->EmbedNormal(
		USER     => $param->{'USER'},
		CRYPT    => $param->{'CRYPT'},
		ROOM     => $param->{'ROOM'},
		FONT     => 'True',
		FONTCTRL => 'True'
	);


	# HTML display
	# Title
	$tmpl->ShowMain($config->{'TagFontTitle'});

	# Table Top
	$table->MakeTop();

	# Cycle through fonts
	while ($res = $statement->fetchrow_hashref())
	{
		# Title
		if ((!defined($type))
		 or ($type ne $res->{'FontType'}))
		{
			# Uppercase
			$type = ucfirst($res->{'FontType'});


			# Display row
			$table->MakeBlank();
			$table->MakeSingle($tmpl->PassSub($type));


			# Alternate for next pass
			$type = $res->{'FontType'};
		}

		# Name handling
		$name = $res->{FontName};

		# Display row
		# Font colour
		if ($type eq 'color')
		{
			# Display colour
			print " <TR>\n",
			      "  <TD><B>$res->{FontName}</B></TD>\n",
			      "  <TD BGCOLOR=\"$res->{FontCode}\" WIDTH=15>&nbsp;</TD>\n",
			      "  <TD><CENTER>", FontFinder($cgi, $param, \@rcolor, $rcolor, $name), "</CENTER></TD>\n",
			      " </TR>\n";
		}

		# Font face
		elsif ($type eq 'face')
		{
			# Current working type
			$ctype = uc($type);

			# Simply show a blank space
			$table->MakeValid(
				"<FONT $ctype=\"$res->{FontCode}\">$res->{FontName}</FONT>",
				FontFinder($cgi, $param, \@rface, $rface, $name)
			 );
		}

		# Font size
		elsif ($type eq 'size')
		{
			# Current working type
			$ctype = uc($type);

			# Simply show a blank space
			$table->MakeValid(
				"<FONT $ctype=\"$res->{FontCode}\">$res->{FontName}</FONT>",
				FontFinder($cgi, $param, \@rsize, $rsize, $name)
			 );
		}
	}

	# Display row
	$table->MakeValid('&nbsp;', "<CENTER>" . $cgi->submit($config->{'TxtSubmitButton'}) . "</CENTER>");

	# End of table
	$table->MakeBottom();
	
	# End query
	$statement->finish();


	# Write new information
	if (defined($param->{'FONTCTRL'}))
	{
		# Evaluation
		eval
		{
			# Create entries
			$rcolor = join(':', @rcolor);
			$rface  = join(':', @rface);
			$rsize  = join(':', @rsize);


			# Update database
			$database->Write("UPDATE Realm
				SET RealmFontColor=?,
				    RealmFontFace=?,
				    RealmFontSize=?
				WHERE RealmName=?",

			 $rcolor,
			 $rface,
			 $rsize,
			 $param->{'ROOM'}
			);

			# Audit entry
			$audit->FontAdjust($param->{'ROOM'}, $param->{'ROOM'});


			# Commit
			$database->Commit();
		};

		# If block failed
		if ($@)
		{
			# Issue warning
			warn("Transaction aborted: $@");

			# Undo incomplete changes
			$database->Rollback();
		}
	}
}

#####################
# Forcasts
# 
# Allows for the host to select a weather station from which the room can have
# weather based upon.   This is the four character ICAO indicator used for 
# METAR reports.

sub Forcasts
{
	####################
	# Data members

	my $database = shift;					# Database handle
	my $cgi      = shift;					# Common gateway interface
	my $param    = shift;					# Parameter list
	my $table    = shift;					# Table handle
	my $tmpl     = shift;					# Template handler
	my $config   = shift;					# Config hash

	my $icoa;						# Station
	my $metar;						# Forcasts


	#####################
	# Program area

	# Retrieval
	($icoa, $metar) = $database->DataGet("SELECT
		  RealmICAO,
		  RealmMETAR
		FROM  Realm
		WHERE RealmName=?",

	 $param->{'ROOM'}
	);


	# Make sure they are defined
	$icoa  = (defined($icoa))  ? $icoa  : 'N/A';
	$metar = (defined($metar)) ? $metar : 'N/A';


	# HTML Display
	# Title
	$tmpl->ShowMain($config->{'TagForcastTitle'});

	# Star form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";

	# Table top
	$table->MakeTop();

	# Forcast
	$table->MakeSingle($tmpl->PassSub($config->{'TagForcastData'}));
	$table->MakeSingle($metar);
	$table->MakeBlank();

	# Selected station
	$table->MakeSingle($tmpl->PassSub($config->{'TagForcastStation'}));
	$table->MakeSingle($cgi->textfield('STATION', $icoa, 4, 4));
	$table->MakeBlank();

	# Submit and bottom
	$table->MakeValid($cgi->submit($config->{'TxtSubmitAlter'}), $cgi->reset($config->{'TxtReset'}));
	$table->MakeBottom();

	# End of form
	print $cgi->end_form();	
}

#####################
# Gummy Handler
#
# Handles the addition and removal of gummies for realm level events.  Making
# it possible for hosts to provide more material to it's patrons outside of
# the norm.

sub GummyHandler
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter hash
	my $table    = shift;					# Table handler
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash

	my $gummies;						# Widget for gummy list
	my $types;						# Widget for types

	my @gummies;						# List of gummies
	my @types    = qw(Canon Mod Super);			# Types

	#####################
	# Program area


	# Listing and display
	# Retieve list
	$database->GetList(\@gummies, "SELECT UPPER(GummyClass) || ': ' || GummyName
		FROM Gummy
		WHERE RealmName=?
		ORDER BY GummyClass, GummyName",

	 $param->{'ROOM'}
	);
	
	# Append safety
	unshift(@gummies, $config->{'TxtNone'});


	# Generate widget
	$gummies = $cgi->scrolling_list('GUMMYREM',  \@gummies, $gummies[0], 10);
	$types   = $cgi->radio_group('GUMMYTYPE', \@types, $types[0], 'true');


	# HTML
	# Title
	$tmpl->ShowMain($config->{'TagLocaleTitle'});

	# Beginning of form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), GUMMIES=>'True'), "\n";

	# Table
	$table->MakeTop();

	# Removal
	$table->MakeSingle($tmpl->PassTitle($config->{'TagGummyRem'}));
	$table->MakeValid($gummies, $cgi->submit($config->{'TxtSubmitDelete'}), 'middle');
	$table->MakeBlank();

	# Addition
	$table->MakeSingle($tmpl->PassTitle($config->{'TagLocaleAddition'}));
	$table->MakeValid(
		$tmpl->PassSub($config->{'TagGummyName'}),
		$cgi->textfield('GUMMYNAME', '', 15, 20)
	 );
	$table->MakeValid(
		$tmpl->PassSub($config->{'TagGummyPath'}),
		$cgi->textfield('GUMMYPATH', '', 30, 200)
	 );

	$table->MakeValid($tmpl->PassSub($config->{'TagGummyType'}), $types);
	$table->MakeValid('&nbsp;', $cgi->submit($config->{'TxtSubmitAddition'}));
	$table->MakeBottom();

	# End of form
	print $cgi->end_form(), "\n";
}

#####################
# Locale Handler
#
# Handles the addition and removal of the locale settings for realms.  This
# will only apply if the realm actually selected strict mode.

sub LocaleHandler
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter hash
	my $table    = shift;					# Table handler
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash

	my $locales;						# Locales widget
	my @locales;						# List of locales


	#####################
	# Program area
	
	# Fetch locale setting
	($verdict) = $database->DataGetRealmLocale($param->{'ROOM'});


	# Allow display
	if ($verdict eq 'strict')
	{
		# Fetch locales
		$database->GetListRealmLocale(\@locales, $param->{'ROOM'});

		# Append safety
		unshift(@locales, $config->{'TxtNone'});


		# Generate widget
		$locales = $cgi->scrolling_list('LOCLREM',  \@locales, $locales[0], 10);


		# HTML
		# Title
		$tmpl->ShowMain($config->{'TagLocaleTitle'});

		# Beginning of form
		print $cgi->start_form();
		print $param->EmbedNormal($param->Flat(), LOCALE=>'True'), "\n";

		# Table
		$table->MakeTop();

		# Removal
		$table->MakeSingle($tmpl->PassSub($config->{'TagLocaleRem'}));
		$table->MakeValid($locales, $cgi->submit($config->{'TxtSubmitDelete'}), 'middle');
		$table->MakeBlank();

		# Addition
		$table->MakeSingle($tmpl->PassSub($config->{'TagLocaleAddition'}));
		$table->MakeValid($cgi->textfield('LOCLADD', '', 25, 50), $cgi->submit($config->{'TxtSubmitAddition'}));
		$table->MakeBottom();

		# End of form
		print $cgi->end_form(), "\n";
	}

	# Display notive
	else
	{
		$table->MakeTop();
		$table->MakeSingle($config->{'MsgLocale'});
		$table->MakeBottom();
	}
}


#####################
# Realm Select
#
# Basic display and prompt user for a selection on which realm to adminster and
# make modifications of some type to. 

sub RealmSelect
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter hash
	my $table    = shift;					# Table handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash

	my @realms;						# Array of realms
	my $realms;						# Popup menu


	#####################
	# Program area

	# Prepare of drop down
	$database->GetListSupervisorRealms(\@realms, $param->{'USER'});
	$realms = $cgi->popup_menu('ROOM', \@realms);


	# HTML Display
	# Title
	$tmpl->ShowMain($config->{'TagRealmSelect'});

	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";

	# Table top
	$table->MakeTop();

	# Relam selection
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagRealmSelectRealm'}),
	 $realms);
	$table->MakeBlank();

	# Submit and bottom
	$table->MakeSingle($cgi->submit($config->{'TxtSubmitView'}));
	$table->MakeBottom();

	# End of form and embed
	print $cgi->end_form(), "\n";	
}


#####################
# Room Ban
#
# Controls the bannings of certain individuals, this is realm specific.   As a
# result a complete ban can result in a password change that will cause the 
# user to not be able to log on any furthur

sub RoomBan
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter hash
	my $table    = shift;					# Table offer
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash

	my $statement;						# Query statement

	my $bird;						# Singular bird
	my $parole;						# Time till parole
	my $jailbirds;						# Jailbird widget

	my @jailbirds;						# List of jailbird puppets


	#####################
	# program area

	# Retreive list of banned and suspects
	# Prepare
	$database->Pull(\$statement, "SELECT
		 PuppeteerLogin AS \"PuppeteerLogin\",
		 getDate(RestrictionTimestamp) AS \"RestrictionTimestamp\"
		FROM  PuppeteerRestriction
		WHERE RealmName=?", 

	 $param->{'ROOM'}
	);


	# Retreive a list of all banned puppets
	while (($bird, $parole) = $statement->fetchrow())
	{ 
		# Populate array
		push(@jailbirds, "$bird ($parole)");	
	}

	# Finish query
	$statement->finish();


	# Addition
	unshift(@jailbirds, $config->{'TxtNone'});

	# Format lists
	$jailbirds = $cgi->scrolling_list('BANNED',  \@jailbirds, $jailbirds[0], 10);


	# HTML
	$tmpl->ShowMain($config->{'TagBanTitle'});

	# Beginning of form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), BAN=>'True'), "\n";

	$table->MakeTop();
	$table->MakeSingle($tmpl->PassSub($config->{'TagBanRem'}));
	$table->MakeValid($jailbirds, $cgi->submit($config->{'TxtSubmitDelete'}), 'middle');
	$table->MakeBlank();
	$table->MakeBlank();

	$table->MakeSingle($tmpl->PassSub($config->{'TagBanAdd'}));
	$table->MakeValid(
	 $tmpl->Pass('TmplSub',  'MTITLE', $config->{'TagBanAddPuppet'}),
	 $cgi->textfield('SUSPECT', '', 20, 30));
	$table->MakeValid(
	 $tmpl->Pass('TmplSub',  'MTITLE', $config->{'TagBanAddTime'}),
	 $cgi->textfield('PAROLE', 0, 3, 3));

	$table->MakeBlank();
	$table->MakeSingle($tmpl->PassTitle($config->{'TagBanReason'}));
	$table->MakeSingle($cgi->textarea(-name=>'DREASON',-rows=>3,-columns=>50,-wrap=>'OFF',-default=>''));

	$table->MakeSingle($cgi->submit($config->{'TxtSubmitAddition'}));
	$table->MakeBottom();

	# End of form
	print $cgi->end_form(), "\n";
}


#####################
# Room Canon
#
# Allows for the host to add in canon handles, meaning handles that
# are restricted to the realm specifically and probably already exist
# on the system.

sub RoomCanon
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter hash
	my $table    = shift;					# Table handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash

	my $audit;						# Auditing
	my $statement;						# Database statement

	my $name;						# Massaged name
	my $canon;						# Hash of cannon handles
	my %realm;						# Hash for realm tags


	#####################
	# Program area

	# Populate hash
	$database->GetHashRealm(\%realm);

	# Define instance
	$audit = new eRealms::Audit($database, $cgi, $param);


	# Display canon handles
	$database->Pull(\$statement, "SELECT
		 PuppetName     AS \"PuppetName\",
		 PuppeteerLogin AS \"PuppeteerLogin\"
		FROM  Puppet
		WHERE PuppetPool=?
		ORDER BY PuppetName",

	 $param->{'ROOM'}
	);

	# Title
	$tmpl->ShowMain($config->{'TagCanonTitle'});
	


	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat());

	# Table top
	$table->MakeTop();
	$table->MakeSingle($tmpl->PassTitle($config->{'TagCanonList'}));
	$table->MakeBlank();

	# Cycle through
	while ($canon = $statement->fetchrow_hashref())
	{
		# Determin names
		$name =  $canon->{PuppetName};
		$name =~ s/\W//gs;


		# Hide
		if (defined($param->{$name}))
		{
			# Evaluate block
			eval
			{
			 	# Remove canon handle
				$database->Write("DELETE FROM Puppet
					WHERE PuppetName=?
					  AND PuppetPool=?",

				 $canon->{'PuppetName'},
				 $param->{'ROOM'}
				);	

				# Cleanup
				$param->Cleanup($name);

				# Record in audit
				$audit->CanonDel($canon->{'PuppetName'}, $param->{'ROOM'});


				# Commit changes
				$database->Commit();
			};

			# If block failed
			if ($@)
			{
				# Issue warning
				warn("Transaction aborted: $@");

				# Undo incomplete changes
				$database->Rollback();
			}
		} 	

		# Display
		else
		{
			# Standard table
			$table->MakeValid("<B>$canon->{PuppetName}</B> ($canon->{PuppeteerLogin})", $cgi->checkbox(-name=>$name, -label=>''));
		}
	}

	# Kill button
	$table->MakeValid('&nbsp;', $cgi->submit($config->{'TxtSubmitDelete'}));

	# End of form
	print $cgi->end_form();

	# End query
	$statement->finish();


	# Spacer
	$table->MakeBlank();
	$table->MakeBlank();
	$table->MakeBlank();


	# Puppet Addition
	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat());


	# HTML Display
	# Table header
	$table->MakeSingle($tmpl->PassTitle($config->{'TagCanonAddition'}));
	$table->MakeBlank();

	
	# Widgets
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagCanonPuppet'}),
	 $cgi->textfield('CCHAR', '', 15, 30));	
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagCanonPuppeteer'}),
	 $cgi->textfield('CUSER', '', 15, 10));	
	$table->MakeBlank();
	

	# Table bottom
	$table->MakeValid('&nbsp;', $cgi->submit($config->{'TxtSubmitAddition'}));
	$table->MakeBottom();
	

	# End of form
	print $cgi->end_form();
}



#####################
# Room Design
#
# Simply put, this section will be handling the design aspects of a realm
# this deals with the splash page, the header and footer used o the pages
# as well as modifications to the communications panel.

sub RoomDesign
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter hash
	my $table    = shift;					# Table handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash

	my $fields;						# Useable format
	my $open;						# Open document
	my $text;						# Text area text

	my @fields;						# Fields to edit


	#####################
	# Program area

	# Fetch design
	$database->GetListDesign(\@fields);


	# Assign value
	$open = (defined($param->{'DFILD'})) ? $param->{'DFILD'} : 'RealmInfo';

	# Generate widget
	$fields = $cgi->popup_menu('DFILD', \@fields, $open); 


	# Fetch  Document
	$text = $database->FetchRealm($open, $param->{'ROOM'});

	# Cleanup
	$param->Cleanup('DFILD','DTEXT');


	# HTML display
	# Title
	$tmpl->ShowMain($config->{'TagDesignTitle'});

	# Document selection
	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), 'DESIGN', 'True'), "\n";	

	# Seperator bar
	$table->MakeTop();
	$table->MakeSingle($tmpl->PassSub($config->{'TagDesignField'}));
	$table->MakeValid($fields, $cgi->submit($config->{'TxtSubmitView'}));
	$table->MakeBlank();
	$table->MakeBlank();


	# Text modification
	# End of form 
	print $cgi->end_form(), "\n";

	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), DFILD=>$open, DESIGN=>'True'), "\n";	

	# Replace macro
	$config->{TagDesignText} =~ s/MDOCUMENT/$open/gs;

	# Seperator bar
	$table->MakeSingle($tmpl->PassSub($config->{TagDesignText}));
	$table->MakeSingle($cgi->textarea(-name=>'DTEXT',-rows=>20,-columns=>70,-wrap=>'OFF',-default=>$text));
	$table->MakeBlank();

	# Submit and bottom
	$table->MakeValid($cgi->submit($config->{'TxtSubmitButton'}), $cgi->submit($config->{'TxtResetButton'}));
	$table->MakeBottom();

	# End of form 
	print $cgi->end_form(), "\n";
}


#####################
# Room Moderators
#
# Allows for the addition of moderators into the system itself.  While not 
# overly necessary as no power is actually gotten from this.  It will
# allow to see more trusted people within the realms.

sub RoomMods
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter hash
	my $table    = shift;					# Table handler
	my $tmpl     = shift;					# Template handler
	my $config   = shift;					# Config hash

	my $option;						# Option handler

	my $acl;						# Access control widget
	my $moderators;						# List of moderator puppets
	my $res;						# Discovered results
	my $statement;						# Queried information

	my $login;						# Returned login
	my $access;						# Access number
	my $trans;						# Translated access


	my @acl;						# Access control list
	my @moderators;						# List of moderator puppets


	#####################
	# program area

	# Create instance
	$option = new eRealms::Option();
	$option->Split(\@acl, $config->{'OptACL'});
	

	# Retreive list of moderators
	$database->Pull(\$statement, "SELECT
		  p.PuppeteerLogin AS \"PuppeteerLogin\",
		  r.RealmACL       AS \"RealmACL\"
                FROM  Puppeteer p, RealmSupervisor r
                WHERE p.PuppeteerLogin = r.PuppeteerLogin
                  AND p.PuppeteerModPrivs='yes'
                  AND r.RealmName=?
		ORDER BY p.PuppeteerLogin",

	 $param->{'ROOM'}
	);

	# Loop and determine
	while (($login, $access) = $statement->fetchrow())
	{
		# Cleanup
		$trans = '';

		# Convoluted system
		if (($access - 8) >= 0) { $access -= 8; $trans = "S$trans"; }
		if (($access - 4) >= 0) { $access -= 4; $trans = "J$trans"; }
		if (($access - 2) >= 0) { $access -= 2; $trans = "D$trans"; }
		if (($access - 1) >= 0) { $access -= 1; $trans = "A$trans"; }

		# Create login name
		$login = "$login ($trans)" if (length($trans) > 0);

		# Append
		push(@moderators, $login);	
	}

	# Finish
	$statement->finish();

	# Append safety measure
	unshift(@moderators, $config->{'TxtNone'});


	# Format widgets
	$moderators = $cgi->scrolling_list('MODREM',  \@moderators, $moderators[0], 10);
	$acl        = $cgi->checkbox_group(-name=>'MODACL', -values=>\@acl, -linebreak=>'true', -default=>$acl[0]);


	# HTML
	# Title
	$tmpl->ShowMain($config->{'TagModTitle'});

	# Beginning of form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), MOD=>'True'), "\n";

	# Table top
	$table->MakeTop();

	# Remove
	$table->MakeSingle($tmpl->PassSub($config->{'TagModRem'}));
	$table->MakeValid($moderators, $cgi->submit($config->{'TxtSubmitDelete'}), 'middle');
	$table->MakeBlank();

	# Addition
	$table->MakeSingle($tmpl->PassTitle($config->{'TagModAdd'}));
	$table->MakeValid(
		$tmpl->PassSub($config->{'TagModLogin'}),
		$cgi->textfield('MODADD', '', 20, 30)
	 );
	$table->MakeValid(
		$tmpl->PassSub($config->{'TagModACL'}),
		$acl
	 );

	# Submit button
	$table->MakeSingle($cgi->submit($config->{'TxtSubmitAddition'}));
	$table->MakeBottom();

	# End of form
	print $cgi->end_form(), "\n";
}


#####################
# Room Picture Ban
#
# This manages the basic display for the addition and removal of pictures from
# the banned list.  Banned implies that the image cannot be used again on the
# system.

sub RoomPicture
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $table    = shift;					# Table handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash

	my $pics;						# List of pics
	my @pics;						# List of pics


	#####################
	# program area

	# Retreive list of images
	$database->GetListImageBan(\@pics);
	unshift(@pics, $config->{'TxtNone'});

	# Create widget
	$pics = $cgi->scrolling_list('PICREM', \@pics, $pics[0], 10);


	# HTML
	# Title
	$tmpl->ShowMain($config->{'TagPicTitle'});

	# Beginning of form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), GRAPHS=>'True'), "\n";

	# Table
	$table->MakeTop();

	# Removal
	$table->MakeSingle($tmpl->PassSub($config->{'TagPicRem'}));
	$table->MakeValid($pics, $cgi->submit($config->{'TxtSubmitDelete'}), 'middle');
	$table->MakeBlank();

	# Addition
	$table->MakeSingle($tmpl->PassSub($config->{'TagPicAdd'}));
	$table->MakeValid($cgi->textfield('PICADD', '', 50, 255), $cgi->submit($config->{'TxtSubmitAddition'}));
	$table->MakeBottom();

	# End of form
	print $cgi->end_form(), "\n";
}


#####################
# RoomView
# 
# Will allow the supervisors to view the selected room and make changes as necessary
# and will probably make the changes themselves.

sub RoomView
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter hash
	my $table    = shift;					# Table handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash
	my $admin    = shift;					# Administrative access
 
	my $comm;						# Comm panel handle
	my $option;						# Option handler
	my $statement;						# Database statement

	my $access;						# Access control
	my $banner;						# Banner widget
	my $color;						# Colour selection default
	my $dateloc;						# Location of timestamp
	my $lang;						# Language restriction
	my $locale;						# Locale display
	my $genre;						# Realm genre
	my $rating;						# Realm rating
	my $realm;						# Realm information
	my $tags;						# Tag enforcement
	my $verified;						# Verification enforcement

	my @optcolr;						# Colour choices 
	my @optgnre;						# Genre choices
	my @optrate;						# Rating choices

	my @optaccs = qw(Relaxed Restricted List);		# Access control
	my @optbner = qw(Display Exempt);			# Banner display
	my @optdate = qw(Bottom Top Removed);			# Timestamp location
	my @optlang = qw(Relaxed Restricted);			# Language access
	my @optlocl = qw(None Strict Variable);			# Locale display
	my @opttags = qw(Relaxed Restricted);			# Tag enforcement
	my @optvery = qw(Yes No);				# Tag enforcement


	#####################
	# Program area

	# Create instances
	$comm   = new eRealms::Comm($database, $cgi);
	$option = new eRealms::Option();

	# Pull colours
	@optcolr = $comm->FontSelectColor($param);

	# Pull genre and rating
	$database->GetListGenre(\@optgnre, 'realm');
	$option->Split(\@optrate, $config->{'SetRating'});


	# Populate realm information hashref
	$database->Pull(\$statement, "SELECT
		 RealmName         AS \"RealmName\",
		 RealmHomepage     AS \"RealmHomepage\",
		 RealmDescription  AS \"RealmDescription\",
		 RealmGenre        AS \"RealmGenre\",
		 RealmRating       AS \"RealmRating\",
		 RealmLanguage     AS \"RealmLanguage\", 
		 RealmFilter       AS \"RealmFilter\",
		 RealmAccess       AS \"RealmAccess\", 
		 RealmVerified     AS \"RealmVerified\", 
		 RealmTags         AS \"RealmTags\", 
		 RealmLocale       AS \"RealmLocale\", 
		 RealmBanner       AS \"RealmBanner\", 
		 RealmTextColour   AS \"RealmTextColour\",
		 RealmFramesetRows AS \"RealmFramesetRows\",
		 RealmFramesetCols AS \"RealmFramesetCols\",
		 RealmImageHeight  AS \"RealmImageHeight\", 
		 RealmImageWidth   AS \"RealmImageWidth\",
		 RealmImageSize    AS \"RealmImageSize\",
		 RealmDateLoc      AS \"RealmDateLoc\",
		 RealmDateFormat   AS \"RealmDateFormat\",
		 RealmAltBoard     AS \"RealmAltBoard\",
		 RealmArchive      AS \"RealmArchive\",
		 RealmBoard        AS \"RealmBoard\"
		FROM Realm 
		WHERE RealmName=?",

	 $param->{'ROOM'}
	);

	# Retreive and finish
	$realm = $statement->fetchrow_hashref();
	$statement->finish();

	# Prepare radiobuttons
	$access   = $cgi->radio_group('ACCESS', \@optaccs, ucfirst($realm->{'RealmAccess'}), 'true');
	$banner   = $cgi->radio_group('BANNER', \@optbner, ucfirst($realm->{'RealmBanner'}), 'true');
	$dateloc  = $cgi->radio_group('DATLOC', \@optdate, ucfirst($realm->{'RealmDateLoc'}), 'true');
	$lang     = $cgi->radio_group('LANG',   \@optlang, ucfirst($realm->{'RealmLanguage'}), 'true');
	$locale   = $cgi->radio_group('LOCL',   \@optlocl, ucfirst($realm->{'RealmLocale'}), 'true');
	$tags     = $cgi->radio_group('TAGS',   \@opttags, ucfirst($realm->{'RealmTags'}), 'true');
	$verified = $cgi->radio_group('VERIFY', \@optvery, ucfirst($realm->{'RealmVerified'}), 'true');

	# Different widget
	$color    = $cgi->popup_menu('COLOR',  \@optcolr, $realm->{'RealmTextColour'});
	$genre    = $cgi->popup_menu('GENRE',  \@optgnre, $realm->{'RealmGenre'});
	$rating   = $cgi->popup_menu('RATING', \@optrate, $realm->{'RealmRating'});

        # Assign values
        $realm->{'RealmAltBoard'}    = (defined($realm->{'RealmAltBoard'}))    ? $realm->{'RealmAltBoard'} : 'http://';
        $realm->{'RealmHomepage'}    = (defined($realm->{'RealmHomepage'}))    ? $realm->{'RealmHomepage'} : 'http://';
        $realm->{'RealmBoard'}       = (defined($realm->{'RealmBoard'}))       ? $realm->{'RealmBoard'} : '';
	$realm->{'RealmArchive'}     = (defined($realm->{'RealmArchive'}))     ? $realm->{'RealmArchive'} : '';
	$realm->{'RealmDescription'} = (defined($realm->{'RealmDescription'})) ? $realm->{'RealmDescription'} : '';
	$realm->{'RealmFilter'}      = (defined($realm->{'RealmFilter'}))      ? $realm->{'RealmFilter'} : '****';


	# HTML display
	# Title
	$tmpl->ShowMain($realm->{'RealmName'});

	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), 'MAIN', 'True'), "\n";	

	# Table top
	$table->MakeTop();
	
	# Access control
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagRealmAccess'}),
	 $access);
	$table->MakeBlank();

	# Locale handling
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagRealmLocale'}),
	 $locale);
	$table->MakeBlank();

	# Realm tag control
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagRealmTags'}),
	 $tags);
	$table->MakeBlank();


	# Administrative
	# Banner exemption
	if ($admin eq 'yes')
	{
		$table->MakeValid(
		 $tmpl->PassSub($config->{'TagRealmBanner'}),
		 $banner);
		$table->MakeBlank();

		$table->MakeValid(
		 $tmpl->PassSub($config->{'TagRealmVerified'}),
		 $verified);
		$table->MakeBlank();
	}

	# Spacer 
	$table->MakeBlank();


	# Lnaguage
	# Date locale
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagRealmDateLoc'}),
	 $dateloc);

	# Date format
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagRealmFormat'}),
	 $cgi->textfield('DFORMAT', $realm->{'RealmDateFormat'}, 30, 30));
	$table->MakeBlank();
	$table->MakeBlank();


	# Language
	# Realm Language
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagRealmLanguage'}),
	 $lang);

	# Replacement word
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagRealmFilter'}),
	 $cgi->textfield('FILTER', $realm->{'RealmFilter'}, 10, 10));
	$table->MakeBlank();
	$table->MakeBlank();


	# Default font
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagRealmFontDef'}),
	 $color);
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagRealmGenre'}),
	 $genre);
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagRealmRating'}),
	 $rating);
	$table->MakeBlank();
	$table->MakeBlank();

	# Framesets
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagFrameRow'}),
	 $cgi->textfield('FRAMEROW', $realm->{'RealmFramesetRows'}, 3, 3));
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagFrameCol'}),
	 $cgi->textfield('FRAMECOL', $realm->{'RealmFramesetCols'}, 3, 3));
	$table->MakeBlank();
	$table->MakeBlank();


	# Image restriction
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagPupImgHeight'}),
	 $cgi->textfield('IMGHEIGHT', $realm->{'RealmImageHeight'}, 3, 3));
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagPupImgWidth'}),
 	 $cgi->textfield('IMGWIDTH',  $realm->{'RealmImageWidth'}, 3, 3));
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagPupImgSize'}),
 	 $cgi->textfield('IMGSIZE',  $realm->{'RealmImageSize'}, 3, 3));
	$table->MakeBlank();
	$table->MakeBlank();


	# Realm homepage
	$table->MakeSingle($tmpl->PassTitle($config->{'TagRealmHomepage'}));
	$table->MakeSingle($cgi->textfield('HOMEPAGE', $realm->{'RealmHomepage'}, 70, 255));
	$table->MakeBlank();

	# Sample code
	if ($realm->{'RealmHomepage'} !~ /^http:\/\/$/)
	{
		# Retrieve
		my ($sample) = $database->DocumentGetWebring();
		my ($ident)  = $cgi->escape($realm->{'RealmName'});


		# Manipulate and escape
		$sample =  $cgi->escapeHTML($sample);
		$sample =~ s/MREALM/$ident/gs;
		$sample =~ s/\n/<BR>/gs;

		# Display
		$table->MakeSingle($tmpl->PassSub($config->{'TagRealmSample'}));
		$table->MakeSingle($sample);
		$table->MakeBlank();
		$table->MakeBlank();
	}


	# Realm description
	$table->MakeSingle($tmpl->PassTitle($config->{'TagRealmDescription'}));
	$table->MakeSingle($cgi->textarea(-name=>'RDESCR',-rows=>3,-columns=>51,-wrap=>'OFF',-default=>$realm->{'RealmDescription'}));
	$table->MakeBlank();

	# Archive Email address
	$table->MakeSingle($tmpl->PassTitle($config->{'TagRealmArchive'}));
	$table->MakeSingle($cgi->textfield('ARCHIVE', $realm->{'RealmArchive'}, 70, 45));
	$table->MakeBlank();

	# Board administrative address
	$table->MakeSingle($tmpl->PassTitle($config->{'TagRealmBoard'}));
	$table->MakeSingle($cgi->textfield('BOARDS', $realm->{'RealmBoard'}, 70, 45));
	$table->MakeBlank();

	# Alternate message board
	$table->MakeSingle($tmpl->PassTitle($config->{'TagRealmAltBoard'}));
	$table->MakeSingle($cgi->textfield('ALTBOARD', $realm->{'RealmAltBoard'}, 70, 255));
	$table->MakeBlank();

	# Submit and bottom
	$table->MakeValid($cgi->submit($config->{'TxtSubmitButton'}), $cgi->submit($config->{'TxtResetButton'}));
	$table->MakeBottom();

	# End of form 
	print $cgi->end_form(), "\n";
}


#####################
# Room Warn
#
# Will list who is in a realm, and allow a moderator or supervisor to send a
# warning that will be recorded.  This allows hosts to see if someone is causing
# a bit of mischeif.

sub RoomWarn
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $table    = shift;					# Table handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash

	my $who;						# Formatted widget
	my @whos;						# Who is online


	#####################
	# program area

	# Retreive list of banned and suspects
	# Manual due to unique purpose
	$database->GetList(\@whos, "SELECT PuppeteerLogin
		FROM Post
		WHERE RealmName=?
		  AND PostTimestamp > getInterval('10 Minutes')
		  AND PuppetName IS NOT NULL",

	 $param->{'ROOM'});

	# Append for safety
	unshift(@whos, $config->{'TxtNone'});

	# Create widget
	$who = $cgi->scrolling_list('WLOGIN',  \@whos, $whos[0], 10);


	# HTML
	# Title
	$tmpl->ShowMain($config->{'TagWarnTitle'});

	# Beginning of form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), WARN=>'True'), "\n";

	# Table Top
	$table->MakeTop();

	# Remove
	$table->MakeSingle($tmpl->PassSub($config->{'TagWarnSelect'}));
	$table->MakeValid($who, $cgi->submit($config->{'TxtWarn'}), 'middle');
	$table->MakeBlank();

	# Addition
	$table->MakeSingle($tmpl->PassSub($config->{'TagWarnMsg'}));
	$table->MakeSingle($cgi->textarea(
		-name    => 'WMSG',
		-default => $config->{'TxtWarnMsg'},
		-rows    => 3,
		-columns => 50,
		-wrap    => 'SOFT')
	 );
	$table->MakeBottom();

	# End of form
	print $cgi->end_form();
}

#####################
# Room Word
#
# Controls word restrictions, as a result one can add or remove words for which
# such words will be replaced from posts by ****.  Simplistic but effective.

sub RoomWord
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $table    = shift;					# Table handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash

	my $words;						# List of words
	my @words;						# List of words	


	#####################
	# program area

	# Retreive list of banned and suspects
	$database->GetListWords(\@words, $param->{'ROOM'});
	unshift(@words, $config->{'TxtNone'});

	# Create widget
	$words = $cgi->scrolling_list('WORDREM',  \@words, $words[0], 10);


	# HTML
	# Title
	$tmpl->ShowMain($config->{'TagWordTitle'});

	# Beginning of form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), WORDS=>'True'), "\n";

	# Table Top
	$table->MakeTop();

	# Remove
	$table->MakeSingle($tmpl->PassSub($config->{'TagWordRem'}));
	$table->MakeValid($words, $cgi->submit($config->{'TxtSubmitDelete'}), 'middle');
	$table->MakeBlank();

	# Addition
	$table->MakeSingle($tmpl->PassSub($config->{'TagWordAdd'}));
	$table->MakeValid($cgi->textfield('WORDADD', '', 10, 10), $cgi->submit($config->{'TxtSubmitAddition'}));
	$table->MakeBottom();

	# End of form
	print $cgi->end_form(), "\n";
}


#####################
# Room Write
# 
# Will handle the writing of information gathered from the forms. 

sub RoomWrite
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $config   = shift;					# Config hash
	my $access   = shift;					# Access realms
	my $admin    = shift;					# Administrative access
	my $acl      = shift;					# Access control list

	my $audit;						# Audit handler
	my $filter;						# Filter handle
	my $option;						# Option handle
	my $send;						# Mail sending

	my $acls;						# Specific total for ACL
	my $choice;						# Choice of the matter
	my $count;						# Counter
	my $time;						# Time values

	my $gummy;						# Gummy name
	my $puppet;						# Puppet name
	my $puppeteer;						# Puppeteers name

	my %system;						# System hash


	#####################
	# Program area
	
	# Audit handle
	$audit  = new eRealms::Audit($database, $cgi, $param);
	$filter = new eRealms::Filter();

	# Evaluation
	eval
	{
		# Word restrictions
		if (defined($param->{'WORDS'}))
		{
			# Removal
			if ((defined($param->{$config->{'TxtSubmitDelete'}}))
			 && ($param->{'WORDREM'} ne $config->{'TxtNone'}))
			{ 
				# Remove entry
				$database->Write("DELETE FROM Restriction 
					WHERE RealmName=?
					  AND RestrictionType='word'
					  AND RestrictionValue=?", 

				 $param->{'ROOM'},
				 $param->{'WORDREM'}
				);

				# Audit entry
				$audit->WordRem($param->{'WORDREM'}, $param->{'ROOM'});
			}

			# Addition
			elsif (defined($param->{$config->{'TxtSubmitAddition'}}))
			{ 
				# Only if a word
				if (($param->{'WORDADD'} !~ /^\W+/sg)
				 && (length($param->{'WORDADD'}) > 2))
				{
					# Create new entry
					$database->Write("INSERT INTO Restriction 
						(RealmName,
						 RestrictionType,
						 RestrictionValue)
						VALUES (?,'word',?)",

				 	 $param->{'ROOM'},
				 	 $param->{'WORDADD'}
					);

					# Audit entry
					$audit->WordAdd($param->{'WORDADD'}, $param->{'ROOM'});
				}
			}	


			# Cleanup
			$param->Cleanup('WORDS');
		}


		# Board Topics
		elsif (defined($param->{'TOPICS'}))
		{
			# Removal of posts and topic
			if ((defined($param->{$config->{'TxtSubmitDelete'}}))
			 && (defined($param->{'CONFIRM'})))
			{
				# Remove the topic
				$database->Write("DELETE FROM BoardTopic 
					WHERE RealmName=?
					  AND BoardTopic=?", 

				 $param->{'ROOM'},
				 $param->{'TOPIC'}
				);

				# Audit entry
				$audit->TopicAdd($param->{'TOPIC'}, $param->{'ROOM'});
			}

			# Addition of topics
			elsif ((defined($param->{"$config->{'TxtSubmitAddition'}"}))
			 && ($param->{'TADD'} !~ /^system/i))
			{
				# Verify validity
				if (($param->{'TADD'} !~ /^\W+/sg)
				 && (length($param->{'TADD'}) > 2))
				{
					# Create
					$database->Write("INSERT INTO BoardTopic 
						(BoardTopic,
						 RealmName)
						VALUES (?,?)", 

				  	 $param->{'TADD'},
				 	 $param->{'ROOM'}
					);

					# New home
					$database->Write("UPDATE Board
						SET BoardTopic=?
						WHERE BoardTopic='system'
						  AND RealmName=?",

					 $param->{'TADD'},
					 $param->{'ROOM'}
					);


					# Audit entry
					$audit->TopicAdd($param->{'TADD'}, $param->{'ROOM'});
				}
			}	

			# Editing of words
			elsif (defined($param->{'TEDIT'}))
			{
				# Clear out HTML and truncate
				$param->{'TEDIT'} = $filter->StripHTML($param->{'TEDIT'}); 
				$param->{'TEDIT'} = substr(DBI::neat($param->{'TEDIT'}, 2002), 1, -1);

				# Final definition
				$param->{'TEDIT'} = (length($param->{'TEDIT'}) > 15) ? $param->{'TEDIT'} : undef;


				# Update database
				$database->Write("UPDATE BoardTopic
					SET BoardDescription=?
					WHERE BoardTopic=?
					  AND RealmName=?",

				 $param->{'TEDIT'},
				 $param->{'TOPIC'},
				 $param->{'ROOM'}
				);

				# Audit entry
				$audit->TopicAdd($param->{'TADD'}, $param->{'ROOM'});
			}

			# Cleanup
			$param->Cleanup('CONFIRM','TOPICS');
		}

	
		# Picture restrictions
		elsif (defined($param->{'GRAPHS'}))
		{
			# Removal
			if ((defined($param->{"$config->{'TxtSubmitDelete'}"}))
			 && ($param->{'GRAPH'} ne $config->{'TxtNone'}))
			{ 
				# Remove
				$database->Write("DELETE FROM Restriction 
					WHERE RealmName='system'
					  AND RestrictionType='picture'
					  AND RestrictionValue=?", 

				 $param->{'PICREM'}
				);

				# Audit entry
				$audit->PicRem($param->{'PICREM'}, $param->{'ROOM'});
			}

			# Addition
			elsif (defined($param->{"$config->{'TxtSubmitAddition'}"}))
			{ 
				# Only if a word
				if (length($param->{'PICADD'}) > 10)
				{
					# Ban the image
					$database->Write("INSERT INTO Restriction 
						(RealmName,
						 RestrictionType,
						 RestrictionValue)
						VALUES ('system','picture',?)", 

				 	 $param->{'PICADD'}
					);

					# Clear all instances of image
					$database->Write("UPDATE Puppet
						SET PuppetPic=NULL,
						    PuppetPicHeight=NULL,
						    PuppetPicWidth=NULL
						WHERE PuppetPic ILIKE '$param->{PICADD}%\'");


					# Audit entry
					$audit->PicAdd($param->{'PICADD'}, $param->{'ROOM'});
				}
			}	

			# Cleanup
			$param->Cleanup('PICS');
		}

		# For banning handling
		elsif (defined($param->{'BAN'}))
		{
			# Removal of person
			if ((defined($param->{"$config->{'TxtSubmitDelete'}"}))
			 && ($param->{'BANNED'} ne $config->{'TxtNone'}))
			{
				# Strip date information
				$param->{'BANNED'} =~ s/\s\(\d{2}-\w{3}-\d{4}\)//s;
				
				# Proper formatting
				($puppeteer) = $database->DataGetPuppeteerExistence($param->{'BANNED'});
			
				# Remove if he is a valid user
			 	if (defined($puppeteer))
				{	
					# Remove
					$database->Write("DELETE FROM PuppeteerRestriction 
						WHERE PuppeteerLogin=?
						  AND RealmName=?", 

					 $puppeteer,
					 $param->{'ROOM'}
					);
	
					# Audit entry
					$audit->BanRem($puppeteer, $param->{'ROOM'});
				}
			}

			# Addition
			elsif (defined($param->{"$config->{'TxtSubmitAddition'}"}))
			{
				# Determine existence
				($puppeteer) = $database->DataGetPuppeteerExistence($param->{'SUSPECT'});

				# Verification
				if ((length($param->{'SUSPECT'}) > 2)
				 && ($param->{'SUSPECT'} !~ /^\s/)
				 && (defined($puppeteer)))
				{
					# Make sure the timestamp is valid
					$param->{'PAROLE'} = ($param->{'PAROLE'} =~ /\d+/) ? $param->{'PAROLE'} : 0;
					$param->{'PAROLE'} = ($param->{'PAROLE'} < 1000)   ? $param->{'PAROLE'} : 1000;


					# Determine when one is paroled
					$time = ($param->{'PAROLE'} < 1) ? "('01/01/3000 00:00:00')::TIMESTAMP" : "(LOCALTIMESTAMP + INTERVAL '$param->{PAROLE} days')::TIMESTAMP";

					# Write and record addition
					$database->Write("INSERT INTO PuppeteerRestriction 
						(PuppeteerLogin,
						 RealmName,
						 RestrictionTimestamp)	
						VALUES (?,?,$time)",

					 $puppeteer, 
					 $param->{'ROOM'}
					); 

					# Email notificiation
					if ($access ne 'list')
					{
						# Link with hash
						$database->GetHashSystem(\%system);

						# Pull values
						my ($banned) = $database->DocumentGetMailBanned();
						my ($email)  = $database->DataGetPuppeteerEmail($puppeteer);
						my ($name)   = $database->DataGetPuppeteerName($puppeteer);
						my ($parole) = $database->DataGet("SELECT getDate($time) AS \"Time\"");

						# Default reason
						$param->{'DREASON'} = (length($param->{'DREASON'}) > 3) ? $param->{'DREASON'} : $system{'SetBannedDefault'};


						# Mail sending
						# Initiate handle
						$send = tie(*MAIL, 'eRealms::Mail', $database);

						# Initial settings
						$send->SetSubject("You have been banned from $param->{ROOM}");
						$send->SetSearch(
							BANDATE    => $parole,
							BANREASON  => $param->{'DREASON'},
							FULLNAME   => $name,
							ROOMNAME   => $param->{'ROOM'},
							SERVERNAME => $system{'SetInfoServer'},
							USERNAME   => $puppeteer
						 );
						$send->ToggleWrap();

						# Recipients
						$send->AddTo("$name <$email>");
						$send->AddBCC("$system{SetBannedName} <$system{SetBannedAddress}>");


						# Message
						print MAIL $banned;

						# Close and send
						close(MAIL);
					}


					# Record in audit
					$audit->BanAdd($puppeteer, $param->{'ROOM'});
				}
			}

			# Cleanup
			$param->Cleanup('BAN','TIMESPAN');
		}


		# Handling of canon handles
		elsif (defined($param->{'CANON'}))
		{
			# Addition
			if ((defined($param->{'CCHAR'})) 
			 && (defined($param->{'CUSER'})) 
			 && (length($param->{'CCHAR'}) > 2))
			{
				# Determine existence
				($puppeteer) = $database->DataGetPuppeteerExistence($param->{'CUSER'});
				($puppet)    = $database->DataGetPuppetExistence($param->{'CCHAR'}, $param->{'CUSER'});


				# Everything in place
				if ((defined($puppeteer))
				 && (!defined($puppet)))
				{
					# Find gummy default
					($gummy) = $database->DataGet("SELECT GummyPath
						FROM Gummy
						WHERE GummyClass=?
						ORDER BY GummyName LIMIT 1", 'canon');


					# Create handle
					$database->Write("INSERT INTO Puppet
						(PuppetName,
						 PuppeteerLogin,
						 PuppetPool,
						 PuppetGummy)
						VALUES(?,?,?,?)",

					 $param->{'CCHAR'},
					 $puppeteer,
					 $param->{'ROOM'},
					 $gummy
					);


					# Audit entry
					$audit->CanonAdd($param->{'CCHAR'}, $param->{'ROOM'});
				}
			}


			# Cleanup
			$param->Cleanup('CDEL','CCHAR','CUSER');
		}


		# Gummy handler
		elsif (defined($param->{'GUMMY'}))
		{
			# Locale additions
			if ((defined($param->{$config->{'TxtSubmitAddition'}}))
			 && (defined($param->{'GUMMYNAME'}))
			 && (defined($param->{'GUMMYPATH'}))
			 && (defined($param->{'GUMMYTYPE'}))
			 && (length($param->{'GUMMYNAME'}) > 2)
			 && (length($param->{'GUMMYPATH'}) > 5))
			{
				# Insert locale
				$database->Write("INSERT INTO Gummy
					(GummyName,
					 GummyClass,
					 GummyPath,
					 RealmName)
					VALUES (?,?,?,?)",

				 $param->{'GUMMYNAME'},
				 lc($param->{'GUMMYTYPE'}),
				 $param->{'GUMMYPATH'},
				 $param->{'ROOM'}
				);

				# Auditing entry
				$audit->LocalGummyAdd($param->{'LOCLADD'}, $param->{'ROOM'});

				# Cleanup
				$param->Cleanup('GUMMYNAME', 'GUMMYTYPE', 'GUMMYPATH');
			}

			# Locale removals
			elsif  ((defined($param->{$config->{'TxtSubmitDelete'}}))
		 	 && ($param->{'GUMMYREM'} ne $config->{'TxtNone'}))
			{
				# Verify for proper string format
				if ($param->{'GUMMYREM'} =~ /([A-Z]+): (.+)/)
				{
					# Pull class andname
					my $gclass = $1;
					my $gname  = $2;


					# Remove locale
					$database->Write("DELETE FROM Gummy
						WHERE GummyName=?
						  AND GummyClass=?
						  AND RealmName=?",

					 $gname,
					 lc($gclass),
					 $param->{'ROOM'}
					);

					# Auditing entry
					$audit->LocalGummyRem($param->{'GUMMYREM'}, $param->{'ROOM'});

					# Cleanup
					$param->Cleanup('GUMMYREM');
				}
			}
		}


		# Local handler
		elsif (defined($param->{'LOCALE'}))
		{
			# Locale additions
			if ((defined($param->{$config->{'TxtSubmitAddition'}}))
			 && (defined($param->{'LOCLADD'}))
			 && (length($param->{'LOCLADD'}) > 2))
			{
				# Insert locale
				$database->Write("INSERT INTO RealmLocale
					(RealmLocale,
					 Realmname)
					VALUES (?,?)",

				 $param->{'LOCLADD'},
				 $param->{'ROOM'}
				);

				# Auditing entry
				$audit->LocaleAdd($param->{'LOCLADD'}, $param->{'ROOM'});

				# Cleanup
				$param->Cleanup('LOCLADD');
			}

			# Locale removals
			elsif  ((defined($param->{$config->{'TxtSubmitDelete'}}))
		 	 && ($param->{'LOCLADD'} ne $config->{'TxtNone'}))
			{
				# Remove locale
				$database->Write("DELETE FROM RealmLocale
					WHERE RealmLocale=?
					  AND RealmName=?",

				 $param->{'LOCLREM'},
				 $param->{'ROOM'}
				);

				# Auditing entry
				$audit->LocaleRem($param->{'LOCLREM'}, $param->{'ROOM'});

				# Cleanup
				$param->Cleanup('LOCLREM');
			}
		}


		# Handling of design
		elsif ((defined($param->{'DESIGN'}))
		    && (defined($param->{$config->{'TxtSubmitButton'}})))
		{
			# Determine Realm Information
			if (length($param->{'DTEXT'}) > 15)
			{
				# Verify for existence
				($choice) = $database->DataGet("SELECT RealmName
					FROM RealmDesign
					WHERE RealmName=?
					  AND TagName=?", 

				 $param->{'ROOM'},
				 $param->{'DFILD'}
				);
			

				# Update
				if (defined($choice))
				{
					# Update information
					$database->Write("UPDATE RealmDesign
						SET DesignContent=?
						WHERE RealmName=?
						  AND TagName=?",

					 $param->{'DTEXT'},
					 $param->{'ROOM'}, 
 					 $param->{'DFILD'}
					);
				}

				# Create instance
				else
				{
					# Insert information
					$database->Write("INSERT INTO RealmDesign
						(DesignContent,
						 TagName,
						 RealmName)
						VALUES (?,?,?)",

					 $param->{'DTEXT'},
					 $param->{'DFILD'},
					 $param->{'ROOM'}
					);
				}
			}

			# Delete entry
			else
			{
				# Remove row and use defaults
				$database->Write("DELETE FROM RealmDesign
					WHERE RealmName=?
					  AND TagName=?",

				 $param->{'ROOM'},
				 $param->{'DFILD'}
				);
			}

			# Audit entry
			$audit->InfoChange($param->{'DFILD'}, $param->{'ROOM'});

			# Cleanup
			$param->Cleanup('DTEXT');
		}

		# For moderator handling
		elsif (defined($param->{'MOD'}))
		{
			# Removal of person
			if ((defined($param->{$config->{'TxtSubmitDelete'}}))
			 && ($param->{'MODREM'} ne $config->{'TxtNone'}))
			{
				# Strip additional material
				$param->{'MODREM'} =~ s/^(.*)\s\(\w+\)$/$1/;

	
				# Initial handle
				# Remove from realm
				$database->Write("DELETE FROM RealmSupervisor
					WHERE PuppeteerLogin=?
					  AND RealmName=?",

				 $param->{'MODREM'},
				 $param->{'ROOM'}
				);

				# Audit entry
				$audit->ModsRem($param->{'MODREM'}, $param->{'ROOM'});


				# Retreive positions
				($count) = $database->DataGetModCount($param->{'MODREM'});

				# Moderator bit handling
				if ($count < 1)
				{
					# Update database
					$database->Write("UPDATE Puppeteer 
						SET   PuppeteerModPrivs='no' 
						WHERE PuppeteerLogin=?", 

					 $param->{'MODREM'}
					);

					# Reset gummy
					$database->Write("UPDATE Puppet 
						SET   PuppetGummy=NULL 
						WHERE PuppeteerLogin=?",

					 $param->{'MODREM'}
					);
				}
			}

			# Addition
			elsif (defined($param->{$config->{'TxtSubmitAddition'}}))
			{

				# Verification
				if ((length($param->{'MODADD'}) > 2) && ($param->{'MODADD'} !~ /^\s/))
				{
					# Retreive supervisory bit
					($choice) = $database->DataGetSuper($param->{'MODADD'});

					# Only if not a supervisor
					if ($choice ne 'yes')
					{
						# Retrieve positions
						($count) = $database->DataGetModCount($param->{'MODADD'});


						# Determine ACL number
						# Create instance
						$option = new eRealms::Option();

						# Initial assignment
						$acls = 0;

						# Determine
						foreach $choice (@{$acl})
						{
							# Determine number and add
							$acls += $option->FindACL($config->{'OptACL'}, $choice);
						}


						# Change moderator bit
						if ($count < 1)
						{
							# Change host bit
							$database->Write("Update Puppeteer 
								SET   PuppeteerModPrivs='yes' 
								WHERE PuppeteerLogin=?",

							 $param->{'MODADD'}
							);
						}
				
						# Add room specific entry
						$database->Write("INSERT INTO RealmSupervisor
							(PuppeteerLogin,
							 RealmName,
							 RealmACL)
							VALUES(?,?,?)",
 
						 $param->{'MODADD'},
						 $param->{'ROOM'},
						 $acls);

						# Record in audit
						$audit->ModsAdd($param->{'MODADD'}, $param->{'ROOM'});
					}
				}
			}

			# Cleanup
			$param->Cleanup('MOD','MODACL');
		}

		# Update main information
		elsif (defined($param->{'MAIN'}))
		{
			# Determine validity
			$param->{'ARCHIVE'}  = ($param->{'ARCHIVE'} =~ /^\w(\w|\.|\-)*?\@(\w|\-)*?\..*?/s) ? $param->{'ARCHIVE'} : undef;
			$param->{'BOARDS'}   = ($param->{'BOARDS'}  =~ /^\w(\w|\.|\-)*?\@(\w|\-)*?\..*?/s) ? $param->{'BOARDS'} : undef;
			$param->{'ALTBOARD'} = ((length($param->{'ALTBOARD'}) > 12) && ($param->{'ALTBOARD'} =~ /^http:\/\//is)) ? $param->{'ALTBOARD'} : undef;
			$param->{'HOMEPAGE'} = ((length($param->{'HOMEPAGE'}) > 12) && ($param->{'HOMEPAGE'} =~ /^http:\/\//is)) ? $param->{'HOMEPAGE'} : undef;
			$param->{'FILTER'}   = (length($param->{'FILTER'}) > 1) ? $param->{'FILTER'} : '****';


			# Realm description
			$param->{'RDESCR'} = $filter->StripHTML($param->{'RDESCR'});
			$param->{'RDESCR'} = substr(DBI::neat($param->{'RDESCR'}, 252), 1, -1);
			$param->{'RDESCR'} = (length($param->{'RDESCR'}) > 50) ? $param->{'RDESCR'} : undef;

			# Massage description
			if (defined($param->{'RDESCR'}))
			{
				# Truncate
				$param->{'RDESCR'} = substr($param->{'RDESCR'}, 0, 250);
		
				# Replace quotes
				$param->{'RDESCR'} =~ s/'/`/gs;
				$param->{'RDESCR'} =~ s/\n/ /gs;
				$param->{'RDESCR'} =~ s/\r/ /gs;
			}


			# Banner exemption
			if ($admin eq 'yes')
			{
				# Special update
				$database->Write("UPDATE Realm
					SET RealmBanner=?,
					    RealmVerified=?
					WHERE RealmName=?",

				 lc($param->{'BANNER'}),
				 lc($param->{'VERIFY'}),
				 $param->{'ROOM'});

				# Cleanup
				$param->Cleanup('BANNER');
			}


			# Ensure numeric
			$param->{'IMGHEIGHT'} = ($param->{'IMGHEIGHT'} =~ /^\d+$/) ? $param->{'IMGHEIGHT'} : 0;
			$param->{'IMGWIDTH'}  = ($param->{'IMGWIDTH'}  =~ /^\d+$/) ? $param->{'IMGWIDTH'}  : 0;
			$param->{'IMGSIZE'}   = ($param->{'IMGSIZE'}   =~ /^\d+$/) ? $param->{'IMGSIZE'}   : 0;

			# Defaults otherwise
			$param->{'FRAMEROW'}  = ($param->{'FRAMEROW'}  =~ /^\d+$/) ? $param->{'FRAMEROW'}  : 115;
			$param->{'FRAMECOL'}  = ($param->{'FRAMECOL'}  =~ /^\d+$/) ? $param->{'FRAMECOL'}  : 65;

			# Font formatting
			$param->{'DFORMAT'}   = (length($param->{'DFORMAT'}) > 5)  ? $param->{'DFORMAT'}  : 'DD-Mon-YYYY, HH24:MI:SS TZ';


			# Update primary information
			$database->Write("UPDATE Realm 
				SET RealmHomepage=?,
				    RealmDescription=?,
				    RealmGenre=?,
				    RealmRating=?,
				    RealmLanguage=?, 
				    RealmFilter=?, 
				    RealmAccess=?, 
				    RealmTags=?,
				    RealmLocale=?,
				    RealmDateLoc=?,
				    RealmDateFormat=?,
				    RealmTextColour=?,
				    RealmFramesetRows=?,
				    RealmFramesetCols=?,
				    RealmImageHeight=?, 
				    RealmImageWidth=?,
				    RealmImageSize=?,
				    RealmArchive=?,
				    RealmBoard=?,
				    RealmAltBoard=?
				WHERE RealmName=?",

			 $param->{'HOMEPAGE'},
			 $param->{'RDESCR'},
			 $param->{'GENRE'},
			 $param->{'RATING'},
			 lc($param->{'LANG'}), 
			 $param->{'FILTER'},
			 lc($param->{'ACCESS'}), 
			 lc($param->{'TAGS'}),
			 lc($param->{'LOCL'}),
			 lc($param->{'DATLOC'}),
			 $param->{'DFORMAT'},
			 $param->{'COLOR'},
			 $param->{'FRAMEROW'},
			 $param->{'FRAMECOL'},
			 $param->{'IMGHEIGHT'}, 
			 $param->{'IMGWIDTH'},
			 $param->{'IMGSIZE'},
			 $param->{'ARCHIVE'},
			 $param->{'BOARDS'},
			 $param->{'ALTBOARD'},
			 $param->{'ROOM'});

			# Audit entry
			$audit->RealmDetails($param->{'ROOM'}, $param->{'ROOM'});
			
			# Cleanup
			$param->Cleanup('MAIN');
		}

		# Weather forcasts
		elsif (defined($param->{'FORCAST'}))
		{
			if (defined($param->{$config->{'TxtSubmitAlter'}}))
			{
				# Data Members
				my $metar;
				my $station = $param->{'STATION'};


				# Assign values
				if ($station =~ /^\w{4}/)
				{
					# Convert to uppercase
					$station = uc($station);

					# Find Forcast
					my $forc = new eRealms::Weather();
					($station, $metar) = $forc->Get($station);

				}
				else
				{
					# Assign to undefined
					$metar   = undef;
					$station = undef;
				}


				# Modification
				# Change database entry
				$database->Write("UPDATE Realm
					SET RealmICAO=?,
					    RealmMETAR=?
					WHERE RealmName=?",

				 $station,
				 $metar,
				 $param->{'ROOM'}
				);

				# Record audit change
				$audit->Forcast($station, $param->{'ROOM'});	
			}
		}

		# Warnings
		elsif (defined($param->{'WARN'}))
		{
		 	# There is a point to do this?		
			if ((defined($param->{'WLOGIN'}))
			 && ($param->{'CHAR'} ne $config->{'TxtNone'}))
			{
				# Pull notice
				my ($warn) = $database->FetchRealmWarning($param->{'ROOM'});

				# Search and replace
				$warn =~ s/MNOTICE/$param->{'WMSG'}/gs;
				$warn =~ s/MLOGIN/$param->{'WLOGIN'}/gs;


				# Create post
				$database->Write("INSERT INTO Post
					(RealmName,
					 PostTo,
					 PostFullFormat,
					 PostImagelessFormat,
					 PostPartialFormat)
					VALUES (?,?,?,?,?)",
				
				 $param->{'ROOM'},
				 $param->{'WLOGIN'},
				 $warn,
				 $warn,
				 $warn);

				# Record audit change
				$audit->Warn($param->{'WLOGIN'}, $param->{'ROOM'});	
			}
		}

		# Commit
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");

		# Undo incomplete changes
		$database->Rollback();
	}
}
